from unittest.mock import AsyncMock

import httpx
import respx
import pytest

from wapitiCore.net.classes import CrawlerConfiguration
from wapitiCore.net import Request
from wapitiCore.net.crawler import AsyncCrawler
from wapitiCore.attack.mod_printer import ModulePrinter


@pytest.mark.asyncio
@respx.mock
async def test_no_printer_detected_via_html():
    respx.get("http://printer.local/").mock(
        return_value=httpx.Response(
            200,
            headers={},
            content="<html><head><title>Unknown Device</title></head><body>No Printer Here</body></html>"
        )
    )
    respx.get(url__regex=r"http://printer.local/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://printer.local/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://printer.local/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModulePrinter(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 0

@pytest.mark.asyncio
@respx.mock
async def test_hp_model_detection_from_html():
    # Mock the root endpoint with no Server header but model info in HTML
    respx.get("http://printer.local/").mock(
        return_value=httpx.Response(
            200,
            headers={},  # No Server header
            content="""
                <html>
                    <head><title>Welcome</title></head>
                    <body>
                        <p>Welcome to your printer!</p>
                        <p>Model: HP LaserJet Pro MFP M428fdw</p>
                    </body>
                </html>
            """
        )
    )

    # Block other endpoints with 404
    respx.get(url__regex=r"http://printer.local/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://printer.local/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://printer.local/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModulePrinter(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "HP LaserJet Pro MFP M428fdw", "versions": [""], "categories": ["Network Equipment"], "groups": ["Printers"]}'
        )
        assert persister.add_payload.call_args_list[0][1]["module"] == "printer"

@pytest.mark.asyncio
@respx.mock
async def test_hp_laserjet_detection():
    respx.get("http://printer.local/").mock(
        return_value=httpx.Response(
            200,
            headers={"Server": "HP LaserJet 400 M401dne"},
            content="<html><head><title>HP LaserJet 400 M401dne</title></head><body>HP Printer</body></html>"
        )
    )
    respx.get(url__regex=r"http://printer.local/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://printer.local/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://printer.local/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModulePrinter(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "HP LaserJet 400 M401dne", "versions": [""], "categories": ["Network Equipment"], "groups": ["Printers"]}'
        )
        assert persister.add_payload.call_args_list[0][1]["module"] == "printer"

@pytest.mark.asyncio
@respx.mock
async def test_hp_laserjet_detection_with_firmware_version():
    respx.get("http://printer.local/").mock(
        return_value=httpx.Response(
            200,
            headers={"Server": "HP LaserJet 400 M401dne"},
            content="<html><head><title>HP LaserJet 400 M401dne</title></head><body>HP Printer</body></html>"
        )
    )

    xml_content = """<?xml version="1.0" encoding="UTF-8"?>
    <prdcfgdyn2:ProductConfigDyn xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:dd="http://www.hp.com/schemas/imaging/con/dictionaries/1.0/" xmlns:prdcfgdyn2="http://www.hp.com/schemas/imaging/con/ledm/productconfigdyn/2009/03/16" xmlns:prdcfgdyn="http://www.hp.com/schemas/imaging/con/ledm/productconfigdyn/2007/11/05" xsi:schemaLocation="http://www.hp.com/schemas/imaging/con/ledm/productconfigdyn/2009/03/16 ../schemas/ledm2/ProductConfigDyn.xsd                               http://www.hp.com/schemas/imaging/con/ledm/productconfigdyn/2007/11/05 ../schemas/ProductConfigDyn.xsd                               http://www.hp.com/schemas/imaging/con/dictionaries/1.0/ ../schemas/dd/DataDictionaryMasterLEDM.xsd">
	<dd:Version>
		<dd:Revision>SVN-IPG-LEDM.690</dd:Revision>
		<dd:Date>2011-03-30</dd:Date>
	</dd:Version>
	<prdcfgdyn:ProductInformation>
		<dd:Version>
			<dd:Revision>V2.12.0</dd:Revision>
			<dd:Date>2022-04-28</dd:Date>
		</dd:Version>
	</prdcfgdyn:ProductInformation>
	</prdcfgdyn2:ProductConfigDyn>"""

    respx.get("http://printer.local/DevMgmt/ProductConfigDyn.xml").mock(
        return_value=httpx.Response(
            200,
            content=xml_content
        )
    )
    respx.get(url__regex=r"http://printer.local/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://printer.local/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://printer.local/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModulePrinter(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "HP LaserJet 400 M401dne", "versions": ["V2.12.0"], "categories": ["Network Equipment"], "groups": ["Printers"]}'
        )
        assert persister.add_payload.call_args_list[0][1]["module"] == "printer"

@pytest.mark.asyncio
@respx.mock
async def test_epson_l6290_detection():
    respx.get("http://printer.local/").mock(
        return_value=httpx.Response(
            200,
            headers={"Server": "EPSON L6290 Series"},
            content="<html><head><title></title></head><body>Home page Printer</body></html>"
        )
    )
    respx.get(url__regex=r"http://printer.local/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://printer.local/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://printer.local/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModulePrinter(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "EPSON L6290 Series", "versions": [""], "categories": ["Network Equipment"], "groups": ["Printers"]}'
        )
        assert persister.add_payload.call_args_list[0][1]["module"] == "printer"

@pytest.mark.asyncio
@respx.mock
async def test_epson_l6290_detection_with_fw_version():
    respx.get("http://printer.local/").mock(
        return_value=httpx.Response(
            200,
            headers={"Server": "EPSON L6290 Series"},
            content="<html><head><title></title></head><body>Home page Printer</body></html>"
        )
    )

    respx.get("http://printer.local/PRESENTATION/HTML/TOP/INDEX.html").mock(
        return_value=httpx.Response(
            200,
            headers={"Server": "EPSON L6290 Series"},
            content='<html><head><title></title></head><body><div class="section main-menu-section-nolink clearfix">'
                    '<h2 class="menu">Firmware Update</h2><p class="guide">Current Version:05.13.XA22P1</p>'
                    '</div></body></html>"'
        )
    )
    respx.get(url__regex=r"http://printer.local/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://printer.local/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://printer.local/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModulePrinter(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "EPSON L6290 Series", "versions": ["05.13.XA22P1"], "categories": ["Network Equipment"], "groups": ["Printers"]}'
        )
        assert persister.add_payload.call_args_list[0][1]["module"] == "printer"



@pytest.mark.asyncio
@respx.mock
async def test_canon_printer_through_html():
    respx.get("http://printer.local/").mock(
        return_value=httpx.Response(
            200,
            headers={"Server": "Canon Web Server"},
            content='<html><head><title></title><p id="deviceType"><span id="deviceName">MF750C Series</span></p></html>'
        )
    )

    respx.get(url__regex=r"http://printer.local/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://printer.local/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://printer.local/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModulePrinter(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "Canon MF750C Series", "versions": [""], "categories": ["Network Equipment"], "groups": ["Printers"]}'
        )
        assert persister.add_payload.call_args_list[0][1]["module"] == "printer"

@pytest.mark.asyncio
@respx.mock
async def test_canon_printer_via_headers():
    respx.get("http://printer.local/").mock(
        return_value=httpx.Response(
            200,
            headers={"Server": "Canon MF633C Caster"},
            content='<html><head><title></title><body>Canon Printer</body></html>'
        )
    )

    respx.get(url__regex=r"http://printer.local/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://printer.local/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://printer.local/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModulePrinter(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "Canon MF633C Caster", "versions": [""], "categories": ["Network Equipment"], "groups": ["Printers"]}'
        )
        assert persister.add_payload.call_args_list[0][1]["module"] == "printer"


@pytest.mark.asyncio
async def test_get_firmware_version_unknown_brand():
    module = ModulePrinter(AsyncMock(), AsyncMock(), {}, AsyncMock())

    result = await module.get_firmware_version("http://printer.local", "canon")

    assert result == ""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the Wapiti project (https://wapiti-scanner.github.io)
# Copyright (C) 2023-2024 Cyberwatch
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
import hashlib
import json
import re
from typing import Optional, Tuple

from httpx import RequestError
from asyncio import TimeoutError
from playwright.async_api import async_playwright, Page, Browser, Error as PlaywrightError

from wapitiCore.parsers.html_parser import Html
from wapitiCore.net import Request
from wapitiCore.net.web import urlparse
from wapitiCore.attack.cms.cms_common import CommonCMS, MSG_TECHNO_VERSIONED, calculate_git_hash
from wapitiCore.net.classes import CrawlerConfiguration
from wapitiCore.net.response import Response
from wapitiCore.definitions.fingerprint_webapp import SoftwareVersionDisclosureFinding
from wapitiCore.definitions.fingerprint import SoftwareNameDisclosureFinding
from wapitiCore.main.log import logging, log_blue

MSG_NO_MAGENTO = "No Magento Detected"


async def playwright_get_content(url: str, page: Page, browser: Browser, crawler_configuration: CrawlerConfiguration) -> str:
    html = ""
    try:
        await page.goto(
            url,
            timeout = crawler_configuration.timeout * 1000,
            wait_until="load"
        )

        html = await page.content()
        return html
    except PlaywrightError as exception:
        logging.exception(exception)
        return ""
    finally:
        await browser.close()
        

async def fetch_source_files(url: str, crawler_configuration: CrawlerConfiguration) -> set:
    my_files_list = set()
    proxy_settings = None
    if crawler_configuration.proxy:
        proxy_settings = {"server": crawler_configuration.proxy}

    try:
        # Initialize a headless browser session to extract source filess
        async with async_playwright() as pw:
            browser = await pw.firefox.launch(
                headless=True,
                proxy=proxy_settings,
                firefox_user_prefs={
                    "network.proxy.allow_hijacking_localhost": True,
                    "devtools.jsonview.enabled": False,
                }
            )
            context = await browser.new_context(
                ignore_https_errors=True, 
                user_agent=crawler_configuration.user_agent
            )
            page = await context.new_page()

            html = await playwright_get_content(url, page, browser, crawler_configuration)
            parsed = Html(html, url)

            # Collect JS and CSS file links
            js_files = {script['src'] for script in parsed.soup.find_all('script', src=True)}
            css_files = {link['href'] for link in parsed.soup.find_all('link', rel="stylesheet", href=True)}

            my_files_list.update(js_files)
            my_files_list.update(css_files)

    except (PlaywrightError, FileNotFoundError, TimeoutError) as exception:
            # Playwright browser may be missing, etc
            logging.exception(exception)
            logging.warning(
                "Could not run headless browser. "
                "Make sure playwright is installed (`pip install playwright`) "
                "and browsers are installed (`playwright install`)."
            )
    except Exception as e:
        logging.error("An error occurred while fetching JS files and CSS files: %s", e)

    return my_files_list

def get_root_url(url):
    # Parse the URL into components
    parsed_url = urlparse(url)
    # Reconstruct the root URL from the scheme and netloc
    root_url = f"{parsed_url.scheme}://{parsed_url.netloc}/"
    return root_url


class ModuleMagentoEnum(CommonCMS):
    """Detect Magento version."""
    PAYLOADS_HASH = "magento2_hash_files.json"
    versions = []
    url_list = []

    def init(self, crawler, persister, options, event, crawler_configuration):
        CommonCMS.__init__(self,
                           crawler=crawler,
                           persister=persister,
                           attack_options=options,
                           crawler_configuration=crawler_configuration)
        self.url_list = []

    async def fetch_source_list(self, url):
        self.url_list = await fetch_source_files(url, self.crawler_configuration)
        return True

    async def check_magento(self, url):
        """Return the list of urls or None if nothing was founded."""
        try:
            request = Request(f'{url}', 'GET')
            response: Response = await self.crawler.async_send(request, follow_redirects=True)
        except RequestError:
            self.network_errors += 1
            return None
        page = Html(response.content, url)
        pattern = re.compile(r'skin/frontend/.*/default/.*')

        keywords = [
            "magento_opensource", "x-magento-init", "Magento_Ui",
            "Magento_PageBuilder", "VarienForm", "Magento_Enterprise"
        ]

        if (response.is_success and
            (
                any(keyword in response.content for keyword in keywords) or
                "Magento" in response.headers.get('X-Powered-By', '')
            )
        ):
            return await self.fetch_source_list(url)

        for script in page.soup.find_all('script', src=True):
            if "Magento_Theme" in script['src']:
                return await self.fetch_source_list(url)

        for link in page.soup.find_all('link', href=True):
            if pattern.search(link['href']):
                return await self.fetch_source_list(url)

        for cookie in response.cookies:
            if cookie.startswith(('frontend', 'X-Magento', 'mage-')):
                return await self.fetch_source_list(url)

        return None

    async def detect_magento_version(self, payloads_hash, hashes):
        versions = {}
        detection_db = self.get_hashes(payloads_hash)
        for content_hash, path in hashes:

            if content_hash and content_hash in detection_db:
                versions[content_hash] = detection_db[content_hash]

        if versions:
            self.versions = set.intersection(*[set(version_list) for version_list in versions.values()])

    async def get_url_hashes(self, url: str) -> Optional[Tuple[str, str]]:
        if get_root_url(url) == await self.persister.get_root_url():
            request = Request(f"{url}", "GET")
            try:
                response: Response = await self.crawler.async_send(request, follow_redirects=True)
            except RequestError:
                self.network_errors += 1
                return "",""

            file_content = response.bytes
            git_hash = calculate_git_hash(file_content)

            return git_hash, url

    async def attack(self, request: Request, response: Optional[Response] = None):
        self.finished = True
        request_to_root = Request(request.url)
        hashes = []
        if await self.check_magento(request_to_root.url):
            for url in self.url_list:
                hash_val = await self.get_url_hashes(url)
                if hash_val:
                    hashes.append(hash_val)
            await self.detect_magento_version(self.PAYLOADS_HASH, hashes)
            self.versions = sorted(self.versions, key=lambda x: x.split('.')) if self.versions else []

            magento_detected = {
                "name": "Magento",
                "versions": list(self.versions),
                "categories": ["CMS Magento"],
                "groups": ["Content"]
            }

            log_blue(
                MSG_TECHNO_VERSIONED,
                "Magento!",
                self.versions
            )

            if self.versions:
                await self.add_info(
                    finding_class=SoftwareVersionDisclosureFinding,
                    request=request_to_root,
                    info=json.dumps(magento_detected),
                )
            await self.add_info(
                finding_class=SoftwareNameDisclosureFinding,
                request=request_to_root,
                info=json.dumps(magento_detected),
            )
        else:
            log_blue(MSG_NO_MAGENTO)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the Wapiti project (https://wapiti-scanner.github.io)
# Copyright (C) 2023-2025 Cyberwatch
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
import json
import re
from typing import Optional
from urllib.parse import urljoin
import xml.etree.ElementTree as ET

from httpx import RequestError

from wapitiCore.attack.attack import Attack
from wapitiCore.definitions.fingerprint import SoftwareNameDisclosureFinding
from wapitiCore.main.log import log_blue, logging
from wapitiCore.net import Request, Response
from wapitiCore.parsers.html_parser import Html


MSG_PRINTER_VERSIONED = "{0} {1} detected"
MSG_NO_PRINTER = "No Printer Detected"
MSG_NO_FIRMWARE = ""  ##@ consider using "No Firmware Detected"
PRINTER_ENDPOINTS = {
    "hp": ["/", "/hp/device/info_deviceStatus.html", "/hp/device/this.LCDispatcher",
           "/hp/device/webAccess/index.html", "/ipp"],
    "epson": ["/", "/PRESENTATION/HTML/TOP/PRTINFO.HTML", "/webconfig",
              "/PRESENTATION/ADVANCED/HTML/PRTINFO.HTML", "/ipp", "/printer"],
    "canon": ["/", "/RemoteUI/", "/status.html", "/DeviceInformation.xml"]
}
PRINTER_BRANDS = PRINTER_ENDPOINTS.keys()


class ModulePrinter(Attack):
    """Detect Printers."""
    name = "printer"

    firmware_version = ""

    async def check_printer(self, url: str, brand: str):
        """Checks if the given URL corresponds to an HP or Epson printer."""
        brand = brand.casefold()

        if brand not in PRINTER_ENDPOINTS:
            return None

        for endpoint in PRINTER_ENDPOINTS[brand]:
            full_url = urljoin(url, endpoint)
            request = Request(full_url, 'GET')
            try:
                response: Response = await self.crawler.async_send(request, follow_redirects=False)
            except RequestError:
                self.network_errors += 1
                continue

            if not response.is_success:
                continue

                # Canon-specific HTML tag match
            if brand == "canon":
                # 1. <span id="deviceName">...</span>
                match = re.search(r'<span\s+id=["\']deviceName["\']>([^<]+)</span>', response.content)
                if match:
                    model = match.group(1).strip()
                    return f"Canon {model}"

                # Check HTTP headers for model info
            server_header = response.headers.get("Server", "")
            if brand in server_header.casefold():
                return server_header

            # Check for model info in HTML
            match = re.search(fr"({brand}\s+\w+[\w\s-]+)", response.content, re.IGNORECASE)
            if match:
                model_name = match.group(1)
                return model_name

        return None

    async def get_firmware_version(self, printer_url: str, brand: str):
        """
        Dispatcher method that routes to the correct parser based on printer brand.
        """
        match brand:
            case "epson":
                firmware = await self._get_firmware_version_epson(printer_url)
            case "hp":
                firmware = await self._get_firmware_version_hp(printer_url)
            case _:
                firmware = MSG_NO_FIRMWARE
        return firmware

    async def _get_firmware_version_epson(self, printer_url: str):
        endpoint_url = urljoin(printer_url, "/PRESENTATION/HTML/TOP/INDEX.html")
        request = Request(endpoint_url, 'GET')

        try:
            response: Response = await self.crawler.async_send(request, follow_redirects=False)
        except RequestError as req_error:
            self.network_errors += 1
            logging.error("Request Error occurred (Epson): %s", req_error)
            return MSG_NO_FIRMWARE

        page = Html(response.content, endpoint_url)
        patterns = [
            r'Current\s*Version[:\s]*([A-Za-z0-9.\-]+)',
            r'Version\s*actuelle[:\s]*([A-Za-z0-9.\-]+)'
        ]

        for p in page.soup.find_all('p'):
            text = p.get_text(separator=' ', strip=True)
            for pattern in patterns:
                match = re.search(pattern, text, re.IGNORECASE)
                if match:
                    return match.group(1)

        return MSG_NO_FIRMWARE

    async def _get_firmware_version_hp(self, printer_url: str):
        endpoint_url = urljoin(printer_url, "/DevMgmt/ProductConfigDyn.xml")
        request = Request(endpoint_url, 'GET')

        try:
            response: Response = await self.crawler.async_send(request, follow_redirects=False)
        except RequestError as req_error:
            self.network_errors += 1
            logging.error("Request Error occurred (HP): %s", req_error)
            return MSG_NO_FIRMWARE

        if not response.is_success:
            return MSG_NO_FIRMWARE

        try:
            xml_content = response.content
            ns = {
                'dd': 'http://www.hp.com/schemas/imaging/con/dictionaries/1.0/',
                'prdcfgdyn': 'http://www.hp.com/schemas/imaging/con/ledm/productconfigdyn/2007/11/05',
            }
            root = ET.fromstring(xml_content)
            revision = root.find('.//prdcfgdyn:ProductInformation/dd:Version/dd:Revision', ns)

            if revision is not None and revision.text:
                return revision.text.strip()
        except ET.ParseError:
            logging.error("Failed to parse XML (HP).")

        return MSG_NO_FIRMWARE


    async def attack(self, request: Request, response: Optional[Response] = None):
        self.finished = True

        for brand in PRINTER_BRANDS:
            try:
                printer_info = await self.check_printer(request.url, brand)
                if printer_info:
                    self.firmware_version = await self.get_firmware_version(request.url, brand)
                    printer_detected = {
                        "name": printer_info,
                        "versions": [self.firmware_version],
                        "categories": ["Network Equipment"],
                        "groups": ["Printers"]
                    }

                    log_blue(MSG_PRINTER_VERSIONED, f"{printer_info}", [self.firmware_version])
                    await self.add_info(
                        finding_class=SoftwareNameDisclosureFinding,
                        request=request,
                        info=json.dumps(printer_detected),
                    )
                    return  # Stop searching if a printer is found
            except RequestError as req_error:
                self.network_errors += 1
                logging.error("Request Error occurred: %s", req_error)

        log_blue(MSG_NO_PRINTER)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the Wapiti project (https://wapiti-scanner.github.io)
# Copyright (C) 2024 Cyberwatch
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
import json
from typing import Optional
from urllib.parse import urljoin

from wapitiCore.parsers.html_parser import Html
from httpx import RequestError

from wapitiCore.attack.network_devices.network_device_common import NetworkDeviceCommon
from wapitiCore.net import Request
from wapitiCore.net.response import Response
from wapitiCore.definitions.fingerprint import SoftwareNameDisclosureFinding
from wapitiCore.main.log import log_blue, logging

MSG_NO_IVANTI_CONNECT = "No Ivanti Connect Secure Detected"
MSG_IVANTI_CONNECT = "Ivanti Connect Secure Detected"
MSG_IVANTI_SERVICE_MANAGER = "Ivanti Service Manager Detected"
MSG_NO_IVANTI_SERVICE_MANAGER = "No Ivanti Service Manager Detected"
MSG_IVANTI_USER_PORTAL = "Ivanti User Portal Detected"
MSG_NO_IVANTI_USER_PORTAL = "No Ivanti User Portal"
MSG_NO_IVANTI = "No Ivanti Product Detected"

class ModuleIvanti(NetworkDeviceCommon):
    """Detect Ivanti."""
    version = []
    device_name = ""

    async def check_ivanti_connect_secure(self, url):
        check_list = ['dana-na/auth/url_default/welcome.cgi','']
        for item in check_list:
            full_url = urljoin(url, item)
            request = Request(full_url, 'GET')
            try:
                response: Response = await self.crawler.async_send(request, follow_redirects=True)
            except RequestError:
                self.network_errors += 1
                raise
            page = Html(response.content, full_url)
            frmLogin = page.soup.find(name="frmLogin")
            if response.is_success and ("Ivanti Connect Secure" in page.title or frmLogin is not None):
                return True
        return False


    async def check_ivanti_service_manager(self, url):
        check_list = ['dana/home/index.cgi',
            'dana-na/auth/url_default/welcome.cgi',
            'dana-na/auth/welcome.cgi',
            '/HEAT/Account/Login',
            'HEAT',
            'HEAT/lib/RespondJs/respond.min.js',
            'lib/RespondJs/respond.min.js',''
        ]
        for item in check_list:
            full_url = urljoin(url, item)
            request = Request(full_url, 'GET')
            try:
                response: Response = await self.crawler.async_send(request, follow_redirects=True)
            except RequestError:
                self.network_errors += 1
                raise
            page = Html(response.content, full_url)
            if response.is_success and "Ivanti Service Manager" in page.title:
                return True
            h1_tag = page.soup.h1
            if h1_tag and response.is_success and "Ivanti Service Manager" in h1_tag.text.strip():
                return True
            if '/lib/RespondJs/respond.min.js' in response.content:
                return True
        return False

    async def check_ivanti_user_portal(self, url):
        check_list = ['mifs/user/login.jsp','']
        for item in check_list:
            full_url = urljoin(url, item)
            request = Request(full_url, 'GET')
            try:
                response: Response = await self.crawler.async_send(request, follow_redirects=True)
            except RequestError:
                self.network_errors += 1
                raise
            page = Html(response.content, full_url)
            if response.is_success and page.title in ["Ivanti User Portal",
                    "Ivanti Portail utilisateur",
                    "Ivanti Admin Portal"]:
                return True
            h1_tag = page.soup.h1
            if h1_tag and response.is_success and "MI_LOGIN_SCREEN" in h1_tag.text.strip():
                return True
        return False

    async def attack(self, request: Request, response: Optional[Response] = None):
        self.finished = True
        request_to_root = Request(request.url)

        try:
            if await self.check_ivanti_connect_secure(request_to_root.url):
                self.device_name = "Ivanti Connect Secure"
                log_blue(MSG_IVANTI_CONNECT)
            elif await self.check_ivanti_service_manager(request_to_root.url):
                self.device_name = "Ivanti Service Manager"
                log_blue(MSG_IVANTI_SERVICE_MANAGER)
            elif await self.check_ivanti_user_portal(request_to_root.url):
                self.device_name = "Ivanti User Portal"
                log_blue(MSG_IVANTI_USER_PORTAL)
            else:
                log_blue(MSG_NO_IVANTI)

            if self.device_name:
                ivanti_detected = {
                    "name": self.device_name,
                    "versions": self.version,
                    "categories": ["Network Equipment"],
                    "groups": ["Content"]
                }
                await self.add_info(
                    finding_class=SoftwareNameDisclosureFinding,
                    request=request_to_root,
                    info=json.dumps(ivanti_detected),
                )

        except RequestError as req_error:
            self.network_errors += 1
            logging.error(f"Request Error occurred: {req_error}")



from urllib.parse import parse_qs
from tempfile import TemporaryDirectory
from shutil import rmtree
from asyncio import Event

import respx
import httpx
import pytest

from wapitiCore.net import Request
from wapitiCore.net.web import urlparse
from wapitiCore.controller.wapiti import Wapiti


@pytest.mark.asyncio
@respx.mock
async def test_resume_crawling():
    stop_event = Event()

    def process(http_request):
        try:
            page = int(parse_qs(urlparse(str(http_request.url)).query)["page"][0])
        except (IndexError, KeyError, ValueError):
            return httpx.Response(200, text="Invalid value")

        if page == 10:
            stop_event.set()

        if page > 20:
            return httpx.Response(200, text="")

        body = "<html><body>"
        body += "<a href='http://perdu.com/?page={0}'>{0}</a>\n".format(page + 1)
        body += "<a href='http://perdu.com/?page={0}'>{0}</a>\n".format(page + 2)
        return httpx.Response(200, text=body)

    respx.get(url__regex=r"http://perdu\.com/$").mock(
        return_value=httpx.Response(200, text="<html><body><a href='http://perdu.com/?page=0'>0</a>")
    )

    respx.get(url__regex=r"http://perdu\.com/\?page=\d+").mock(side_effect=process)

    # Mock HTTP 404 behavior check
    respx.get(url__regex=r"http://perdu\.com/z.*\.html$").mock(return_value=httpx.Response(404))

    temp_obj = TemporaryDirectory()
    wapiti = Wapiti(Request("http://perdu.com/"), session_dir=temp_obj.name)
    await wapiti.init_persister()
    await wapiti.load_scan_state()
    await wapiti.browse(stop_event, parallelism=1)
    await wapiti.save_scan_state()
    remaining_requests = set([request async for request in wapiti.persister.get_to_browse()])
    # Got root url + pages 0 to 9
    all_requests = set([request async for request, __ in wapiti.persister.get_links()])
    remaining_urls = {request.url for request in remaining_requests - all_requests}
    # Page 10 stops the crawling but gave links to pages 11 and 12 so they will be the remaining urls
    assert remaining_urls == {"http://perdu.com/?page=11", "http://perdu.com/?page=12"}

    wapiti = Wapiti(Request("http://perdu.com/"), session_dir=temp_obj.name)
    await wapiti.init_persister()
    await wapiti.load_scan_state()
    await wapiti.browse(stop_event)
    await wapiti.save_scan_state()
    remaining_requests = set([request async for request in wapiti.persister.get_to_browse()])
    all_requests = set([request async for request, __ in wapiti.persister.get_links()])
    # We stop giving new links at page > 20 but page 20 will give urls for 21 and 22
    # so, we have 24 paginated pages (23 from 0 to 22) + root url here
    assert len(all_requests) == 24
    # We are done as we scanned all the pages
    assert not remaining_requests - all_requests
    rmtree(temp_obj.name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the Wapiti project (https://wapiti-scanner.github.io)
# Copyright (C) 2024 Cyberwatch
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
import json
import re
from typing import Optional
from urllib.parse import urljoin

from wapitiCore.parsers.html_parser import Html
from httpx import RequestError

from wapitiCore.attack.attack import Attack
from wapitiCore.net import Request
from wapitiCore.net.response import Response
from wapitiCore.definitions.fingerprint import SoftwareNameDisclosureFinding
from wapitiCore.main.log import log_blue, logging

MSG_NO_CITRIX = "No Citrix Product Detected"
MSG_CITRIX_DETECTED = "{0}{1} Detected !"


class ModuleCitrix(Attack):
    """Detect Citrix Devices."""

    device_name = "Citrix"
    version = []

    async def check_citrix(self, url):
        check_list = ['logon/LogonPoint/', '']

        for item in check_list:
            full_url = urljoin(url, item)
            request = Request(full_url, 'GET')
            try:
                response: Response = await self.crawler.async_send(request, follow_redirects=False)
            except RequestError:
                self.network_errors += 1
                continue

            if response.is_success:
                return await self.detect_citrix_product(response.content, full_url)

        return False

    async def detect_citrix_product(self, response_content, url: str = ""):
        page = Html(response_content, url)
        title_tag = page.soup.title
        # If title tag exists and has a class attribute
        if title_tag:
            if 'class' in title_tag.attrs:
                title_class = title_tag['class']
                # Assuming class is a list, extracting the first class
                if title_class:
                    extract_pattern = r"^_ctxstxt_(.*)$"
                    match = re.search(extract_pattern, title_class[0])
                    if match:
                        self.device_name = match.group(1)
                        return True

                    return False
            else:
                title = title_tag.text
                if "Citrix" in title:
                    # Extract the product name from the title
                    self.device_name = title
                    return True

                if "NetScaler" in title:
                    # Search the product name in the content
                    matches = page.soup.find_all(
                        "span",
                        string=re.compile(r"^(NetScaler ADC|Citrix NetScaler|NetScaler|NetScaler AWS)$")
                    )
                    if matches:
                        # Set product_name to the first matched product name
                        self.device_name = matches[0].text
                    else:
                        self.device_name = "Citrix"
                    return True

        return False

    async def attack(self, request: Request, response: Optional[Response] = None):
        self.finished = True
        request_to_root = Request(request.url)

        try:
            if await self.check_citrix(request_to_root.url):
                citrix_detected = {
                    "name": self.device_name,
                    "versions": self.version,
                    "categories": ["Network Equipment"],
                    "groups": ["Content"]
                }
                log_blue(
                    MSG_CITRIX_DETECTED,
                    self.device_name,
                    self.version
                )

                await self.add_info(
                    finding_class=SoftwareNameDisclosureFinding,
                    request=request_to_root,
                    info=json.dumps(citrix_detected),
                )
            else:
                log_blue(MSG_NO_CITRIX)
        except RequestError as req_error:
            self.network_errors += 1
            logging.error(f"Request Error occurred: {req_error}")



import os
import sys
from os.path import join as path_join
from unittest.mock import AsyncMock, patch

import httpx
import respx
import pytest

from wapitiCore.net.classes import CrawlerConfiguration
from wapitiCore.net import Request
from wapitiCore.net.crawler import AsyncCrawler
from wapitiCore.attack.mod_cms import ModuleCms
from wapitiCore.attack.cms.mod_magento_enum import fetch_source_files as fetch_source_files_magento, get_root_url
from wapitiCore.attack.cms.mod_typo3_enum import fetch_source_files as fetch_source_files_typo3


# Test no Drupal detected
@pytest.mark.asyncio
@respx.mock
async def test_no_drupal():
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Pas de panique, on va vous aider</h2> \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    # Response to check that we have no more false positives
    respx.get("http://perdu.com/core/misc/drupal.js").mock(return_value=httpx.Response(200))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert not persister.add_payload.call_count

@pytest.mark.asyncio
@respx.mock
async def test_drupal_version_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/drupal/")
    changelog_file = "CHANGELOG.txt"

    with open(path_join(test_directory, changelog_file), errors="ignore") as changelog:
        data = changelog.read()

    # Response to tell that Drupal is used
    respx.get("http://perdu.com/core/misc/drupal.js").mock(
        return_value=httpx.Response(
            200,
            headers={"Content-Type": "application/javascript"})
    )

    # Response for changelog.txt
    respx.get("http://perdu.com/CHANGELOG.txt").mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 2
        assert persister.add_payload.call_args_list[0][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[0][1]["category"] == "Fingerprint web application framework"
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "Drupal", "versions": ["7.67"], "categories": ["CMS Drupal"], "groups": ["Content"]}'
        )
        assert persister.add_payload.call_args_list[1][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[1][1]["category"] == "Fingerprint web technology"
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "Drupal", "versions": ["7.67"], "categories": ["CMS Drupal"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_drupal_multi_versions_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/drupal/")
    maintainers_file = "MAINTAINERS.txt"

    with open(path_join(test_directory, maintainers_file), errors="ignore") as maintainers:
        data = maintainers.read()

    # Response to tell that Drupal is used
    respx.get("http://perdu.com/core/misc/drupal.js").mock(
        return_value=httpx.Response(
            200,
            headers={"Content-Type": "application/javascript"}
        )
    )

    # Response for  maintainers.txt
    respx.get("http://perdu.com/core/MAINTAINERS.txt").mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 2
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "Drupal", "versions": ["8.0.0-beta4", "8.0.0-beta5", "8.0.0-beta6"], "categories": ["CMS Drupal"], "groups": ["Content"]}'
        )
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "Drupal", "versions": ["8.0.0-beta4", "8.0.0-beta5", "8.0.0-beta6"], "categories": ["CMS Drupal"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_drupal_version_not_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/drupal/")
    changelog_edited = "CHANGELOG_EDITED.txt"

    with open(path_join(test_directory, changelog_edited), errors="ignore") as changelog:
        data = changelog.read()

    # Response to tell that Drupal is used
    respx.get("http://perdu.com/misc/drupal.js").mock(
        return_value=httpx.Response(
            200,
            headers={"Content-Type": "application/javascript"}
        )
    )

    # Response for edited changelog.txt
    respx.get("http://perdu.com/CHANGELOG.txt").mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "Drupal", "versions": [], "categories": ["CMS Drupal"], "groups": ["Content"]}'
        )

@pytest.mark.asyncio
@respx.mock
async def test_drupal_plugin_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/drupal/")
    changelog_edited = "CHANGELOG_EDITED.txt"

    with open(path_join(test_directory, changelog_edited), errors="ignore") as changelog:
        data = changelog.read()

    # Response to tell that Drupal is used
    respx.get("http://perdu.com/misc/drupal.js").mock(
        return_value=httpx.Response(
            200,
            headers={"Content-Type": "application/javascript"}
        )
    )

    # Response for edited changelog.txt
    respx.get("http://perdu.com/CHANGELOG.txt").mock(return_value=httpx.Response(200, text=data))
    # Response for installed plugin
    respx.get("http://perdu.com/modules/contrib/ctools/").mock(return_value=httpx.Response(403, content="Forbidden"))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 2
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "Drupal", "versions": [], "categories": ["CMS Drupal"], "groups": ["Content"]}'
        )
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "ctools", "versions": [], "categories": ["Drupal Plugin"], "groups": ["Add-ons"]}'
        )
@pytest.mark.asyncio
@respx.mock
async def test_drupal_plugins_403():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/drupal/")
    changelog_edited = "CHANGELOG_EDITED.txt"

    with open(path_join(test_directory, changelog_edited), errors="ignore") as changelog:
        data = changelog.read()

    # Response to tell that Drupal is used
    respx.get("http://perdu.com/misc/drupal.js").mock(
        return_value=httpx.Response(
            200,
            headers={"Content-Type": "application/javascript"}
        )
    )

    # Response for edited changelog.txt
    respx.get("http://perdu.com/CHANGELOG.txt").mock(return_value=httpx.Response(200, text=data))
    # Response for installed plugin
    respx.get("http://perdu.com/modules/contrib/ctools/").mock(return_value=httpx.Response(403, content="Forbidden"))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(403))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "Drupal", "versions": [], "categories": ["CMS Drupal"], "groups": ["Content"]}'
        )
       

@pytest.mark.asyncio
@respx.mock
async def test_no_joomla():
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Pas de panique, on va vous aider</h2> \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    # Response to check that we have no more false positives

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert not persister.add_payload.call_count


@pytest.mark.asyncio
@respx.mock
async def test_joomla_version_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/joomla/")
    joomla_file = "joomla.xml"

    with open(path_join(test_directory, joomla_file), errors="ignore") as joomla:
        data = joomla.read()

    # Response to tell that Joomla is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="<html><head></head><body>This website use Joomla /administrator/ </body></html>")
    )

    # Response for joomla.xml
    respx.get("http://perdu.com/administrator/manifests/files/joomla.xml").mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 2
        assert persister.add_payload.call_args_list[0][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[0][1]["category"] == "Fingerprint web application framework"
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "Joomla!", "versions": ["3.10.12"], "categories": ["CMS Joomla"], "groups": ["Content"]}'
        )
        assert persister.add_payload.call_args_list[1][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[1][1]["category"] == "Fingerprint web technology"
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "Joomla!", "versions": ["3.10.12"], "categories": ["CMS Joomla"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_joomla_multi_versions_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/joomla/")
    helpsite_file = "helpsite.js"

    with open(path_join(test_directory, helpsite_file), errors="ignore") as helpsite:
        data = helpsite.read()

    # Response to tell that Joomla is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="<html><head></head><body>This website use Joomla /administrator/ </body></html>")
    )

    # Response for  maintainers.txt
    respx.get("http://perdu.com/media/system/js/helpsite.js")\
        .mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 2
        assert persister.add_payload.call_args_list[0][1]["category"] == "Fingerprint web application framework"
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "Joomla!", "versions": ["3.3.4", "3.3.5", "3.3.6", "3.4.0-alpha"], "categories": ["CMS Joomla"], "groups": ["Content"]}'
        )
        assert persister.add_payload.call_args_list[1][1]["category"] == "Fingerprint web technology"
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "Joomla!", "versions": ["3.3.4", "3.3.5", "3.3.6", "3.4.0-alpha"], "categories": ["CMS Joomla"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_joomla_version_not_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/joomla/")
    joomla_edited = "joomla_edited.xml"

    with open(path_join(test_directory, joomla_edited), errors="ignore") as joomla:
        data = joomla.read()

    # Response to tell that Joomla is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="<html><head></head><body>This website use Joomla /administrator/ </body></html>")
    )

    # Response for edited changelog.txt
    respx.get("http://perdu.com/administrator/manifests/files/joomla.xml").mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "Joomla!", "versions": [], "categories": ["CMS Joomla"], "groups": ["Content"]}'
        )

@pytest.mark.asyncio
@respx.mock
async def test_no_prestashop():
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Pas de panique, on va vous aider</h2> \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    # Response to check that we have no more false positives

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert not persister.add_payload.call_count


@pytest.mark.asyncio
@respx.mock
async def test_no_magento():
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Pas de panique, on va vous aider</h2> \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    # Response to check that we have no more false positives
    respx.get("http://perdu.com/test/file.js").mock(return_value=httpx.Response(200))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20, "cms": "magento"}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert not persister.add_payload.call_count


@pytest.mark.asyncio
@respx.mock
async def test_magento_version_not_detected():

    # Response to tell that Magento is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="<html><head><title>Vous Etes Perdu ?</title></head><body> \
                        <link href='http://perdu.com/skin/frontend/test/default/test.css'> \
                        <h1>Perdu sur l'Internet ?</h1></body></html>")
    )

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20, "cms": "magento"}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "Magento", "versions": [], "categories": ["CMS Magento"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_magento_via_script_src():
    # Response to tell that Magento is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="<html><head><title>Vous Etes Perdu ?</title></head><body> \
                    <script src='http//perdu.com/resource/Magento_Theme/test.js'></script> \
                    <h1>Perdu sur l'Internet ?</h1></body></html>")
    )

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20, "cms": "magento"}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "Magento", "versions": [], "categories": ["CMS Magento"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_magento_via_cookies():
    # Response to tell that PrestaShop is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="This website uses Magento",
            headers={"Set-Cookie": "X-Magento=testcookie; Path=/; HttpOnly}"})
    )

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20, "cms": "magento"}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "Magento", "versions": [], "categories": ["CMS Magento"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_magento_multi_version_detected():
    # Sample mocked content from the page
    mocked_content = (
        '<html><head><link rel="stylesheet" type="text/css" '
        'href="http://perdu.com/app/code/Magento/Swatches/view/frontend/web/css/swatches.css">'
        '</head><body>This website uses Magento magento_opensource = []</body></html>'
    )

    # Mock async HTTP responses
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(200, content=mocked_content)
    )

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/magento/")
    swatches_file = "swatches.css"

    with open(os.path.join(test_directory, swatches_file), errors="ignore") as swatches:
        data = swatches.read()

    respx.get("http://perdu.com/app/code/Magento/Swatches/view/frontend/web/css/swatches.css").mock(
        return_value=httpx.Response(200, text=data)
    )

    # Mocking browser behavior - mock playwright_get_content to return a mock page content
    with patch("wapitiCore.attack.cms.mod_magento_enum.playwright_get_content", return_value=mocked_content):
        # Mock the persister
        persister = AsyncMock()
        persister.get_root_url.return_value = "http://perdu.com/"

        # Setup request and crawler configuration
        request = Request("http://perdu.com/")
        request.path_id = 1

        crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
        async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
            options = {"timeout": 10, "level": 2, "tasks": 20, "cms": "magento"}
            module = ModuleCms(crawler, persister, options, crawler_configuration)

            # Run the attack
            await module.attack(request)

            # Assertions for added payloads
            assert persister.add_payload.call_count == 2
            assert persister.add_payload.call_args_list[0][1]["category"] == "Fingerprint web application framework"
            assert persister.add_payload.call_args_list[0][1]["info"] == (
                '{"name": "Magento", "versions": ["2.1.10", "2.1.11", "2.1.12"], "categories": ["CMS Magento"], "groups": ["Content"]}'
            )
            assert persister.add_payload.call_args_list[1][1]["category"] == "Fingerprint web technology"
            assert persister.add_payload.call_args_list[1][1]["info"] == (
                '{"name": "Magento", "versions": ["2.1.10", "2.1.11", "2.1.12"], "categories": ["CMS Magento"], "groups": ["Content"]}'
            )


@pytest.mark.asyncio
@respx.mock
async def test_magento_version_detected():
    # Mocked content for the test
    mocked_content = (
        '<html><head><link href="http://perdu.com/skin/frontend/test/default/test.css">'
        '<link rel="stylesheet" type="text/css" '
        'href="http://perdu.com/app/code/Magento/Swatches/view/frontend/web/css/swatches.css">'
        '</head><body>This website uses Magento magento_opensource = []</body></html>'
    )
    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/magento/")
    swatches_file = "swatches2.css"

    with open(path_join(test_directory, swatches_file), errors="ignore") as swatches:
        data = swatches.read()

    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(200, content=mocked_content)
    )
    respx.get("http://perdu.com/app/code/Magento/Swatches/view/frontend/web/css/swatches.css").mock(
        return_value=httpx.Response(200, text=data)
    )
    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    # Mocking browser behavior - mock playwright_get_content to return a mock page content
    with patch("wapitiCore.attack.cms.mod_magento_enum.playwright_get_content", return_value=mocked_content):
        # Mock the persister
        persister = AsyncMock()
        persister.get_root_url.return_value = "http://perdu.com/"

        # Setup request and crawler configuration
        request = Request("http://perdu.com/")
        request.path_id = 1

        crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
        async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
            options = {"timeout": 10, "level": 2, "tasks": 20, "cms": "magento"}
            module = ModuleCms(crawler, persister, options, crawler_configuration)

            # Run the attack
            await module.attack(request)

            assert persister.add_payload.call_count == 2
            assert persister.add_payload.call_args_list[0][1]["category"] == "Fingerprint web application framework"
            assert persister.add_payload.call_args_list[0][1]["info"] == (
                '{"name": "Magento", "versions": ["2.2.8"], "categories": ["CMS Magento"], "groups": ["Content"]}'
            )
            assert persister.add_payload.call_args_list[1][1]["category"] == "Fingerprint web technology"
            assert persister.add_payload.call_args_list[1][1]["info"] == (
                '{"name": "Magento", "versions": ["2.2.8"], "categories": ["CMS Magento"], "groups": ["Content"]}'
            )


@pytest.mark.asyncio
@respx.mock
async def test_fetch_source_files():
    test_url = "http://perdu.com/"

    # Sample HTML content with JS and CSS links
    mocked_content = '''
    <html>
        <head>
            <link rel="stylesheet" type="text/css" href="http://perdu.com/static/styles.css">
            <script src="http://perdu.com/static/script.js"></script>
        </head>
        <body>Hello World</body>
    </html>
    '''

    # Mock HTTP response for the requested page
    respx.get(test_url).mock(return_value=httpx.Response(200, content=mocked_content))

    # Mocking browser behavior - mock playwright_get_content to return a mock page content
    with patch("wapitiCore.attack.cms.mod_magento_enum.playwright_get_content", return_value=mocked_content):
        crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
        # Run the function
        result = await fetch_source_files_magento(test_url, crawler_configuration)

        # Expected extracted files
        expected_files = {
            "http://perdu.com/static/styles.css",
            "http://perdu.com/static/script.js"
        }

        # Assertions
        assert result == expected_files


@pytest.mark.asyncio
async def test_get_root_url():
    url = "http://perdu.com/some/path"
    expected_root = "http://perdu.com/"

    result = get_root_url(url)

    assert result == expected_root


@pytest.mark.asyncio
@respx.mock
async def test_prestashop_version_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/prestashop/")
    prestashop_file = "admin.js"

    with open(path_join(test_directory, prestashop_file), errors="ignore") as prestashop:
        data = prestashop.read()

    # Response to tell that PrestaShop is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="<html><head><title>Helo</title>\
            </head><body>This website uses Prestashop, prestashop = []</body></html>")
        )

    # Response for admin.js
    respx.get("http://perdu.com/js/admin.js")\
        .mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 2
        assert persister.add_payload.call_args_list[0][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[0][1]["category"] == "Fingerprint web application framework"
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "PrestaShop", "versions": ["1.6.0.5"], "categories": ["CMS PrestaShop"], "groups": ["Content"]}'
        )
        assert persister.add_payload.call_args_list[1][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[1][1]["category"] == "Fingerprint web technology"
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "PrestaShop", "versions": ["1.6.0.5"], "categories": ["CMS PrestaShop"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_prestashop_multi_versions_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/prestashop/")
    helpsite_file = "theme.css"

    with open(path_join(test_directory, helpsite_file), errors="ignore") as helpsite:
        data = helpsite.read()

    # Response to tell that PrestaShop is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="This website use PrestaShop, prestashop = []")
    )

    # Response for  theme.css
    respx.get("http://perdu.com/themes/classic/assets/css/theme.css")\
        .mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 2
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "PrestaShop", "versions": ["1.7.7.6", "1.7.7.7", "1.7.7.8"], "categories": ["CMS PrestaShop"], "groups": ["Content"]}'
        )
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "PrestaShop", "versions": ["1.7.7.6", "1.7.7.7", "1.7.7.8"], "categories": ["CMS PrestaShop"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_prestashop_version_not_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/prestashop/")
    prestashop_edited = "theme_edited.css"

    with open(path_join(test_directory, prestashop_edited), errors="ignore") as prestashop:
        data = prestashop.read()

    # Response to tell that PrestaShop is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="This website use Prestashop prestashop = []")
    )

    # Response for edited changelog.txt
    respx.get("http://perdu.com/themes/classic/assets/css/theme.css").mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "PrestaShop", "versions": [], "categories": ["CMS PrestaShop"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_spip_version_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/spip/")
    spip_file = "CHANGELOG.txt"

    with open(path_join(test_directory, spip_file), errors="ignore") as spip:
        data = spip.read()

    # Response to tell that SPIP is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="This website use SPIP, your_spip_attribute = ''",
            headers={"composed-by": "SPIP"})
    )

    # Response for CHANGELOG.txt
    respx.get("http://perdu.com/CHANGELOG.txt").mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 2
        assert persister.add_payload.call_args_list[0][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[0][1]["category"] == "Fingerprint web application framework"
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "SPIP", "versions": ["v3.1.14"], "categories": ["CMS SPIP"], "groups": ["Content"]}'
        )
        assert persister.add_payload.call_args_list[1][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[1][1]["category"] == "Fingerprint web technology"
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "SPIP", "versions": ["v3.1.14"], "categories": ["CMS SPIP"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_spip_multi_versions_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/spip/")
    ajax_file = "ajaxCallback.js"

    with open(path_join(test_directory, ajax_file), errors="ignore") as ajax:
        data = ajax.read()

    # Response to tell that SPIP is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="This website use SPIP, your_spip_attribute = ''",
            headers={"composed-by": "SPIP"})
    )

    # Response for ajaxCallback.js
    respx.get("http://perdu.com/prive/javascript/ajaxCallback.js")\
        .mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 2
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "SPIP", "versions": ["v2.0.0", "v2.0.1", "v2.0.2", "v2.0.3"], "categories": ["CMS SPIP"], "groups": ["Content"]}'
        )
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "SPIP", "versions": ["v2.0.0", "v2.0.1", "v2.0.2", "v2.0.3"], "categories": ["CMS SPIP"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_spip_version_not_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/spip/")
    spip_edited = "CHANGELOG_edited.txt"

    with open(path_join(test_directory, spip_edited), errors="ignore") as spip:
        data = spip.read()

    # Response to tell that SPIP is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="This website use SPIP, your_spip_attribute = ''",
            headers={"composed-by": "SPIP"})
    )

    # Response for edited changelog.txt
    respx.get("http://perdu.com/CHANGELOG.txt").mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "SPIP", "versions": [], "categories": ["CMS SPIP"], "groups": ["Content"]}'
        )
@pytest.mark.asyncio
@respx.mock
async def test_spip_plugins():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/spip/")
    spip_edited = "CHANGELOG_edited.txt"

    with open(path_join(test_directory, spip_edited), errors="ignore") as spip:
        data = spip.read()

    # Response to tell that SPIP is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="This website use SPIP, your_spip_attribute = ''",
            headers={"composed-by": "SPIP"})
    )

    # Response for edited changelog.txt
    respx.get("http://perdu.com/CHANGELOG.txt").mock(return_value=httpx.Response(200, text=data))
    # Responses for plugin folders
    respx.get("http://perdu.com/plugins/oembed/").mock(return_value=httpx.Response(403))
    respx.get("http://perdu.com/plugins-dist/nuage/").mock(return_value=httpx.Response(403))


    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 3

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20, "cms": "spip"}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 3
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "SPIP", "versions": [], "categories": ["CMS SPIP"], "groups": ["Content"]}'
        )
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "nuage", "versions": [], "categories": ["SPIP Plugin"], "groups": ["Add-ons"]}'
        )
        assert persister.add_payload.call_args_list[2][1]["info"] == (
            '{"name": "oembed", "versions": [], "categories": ["SPIP Plugin"], "groups": ["Add-ons"]}'
        )

@pytest.mark.asyncio
@respx.mock
async def test_spip_no_plugins_403():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/spip/")
    spip_edited = "CHANGELOG_edited.txt"

    with open(path_join(test_directory, spip_edited), errors="ignore") as spip:
        data = spip.read()

    # Response to tell that SPIP is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="This website use SPIP, your_spip_attribute = ''",
            headers={"composed-by": "SPIP"})
    )

    # Response for edited changelog.txt
    respx.get("http://perdu.com/CHANGELOG.txt").mock(return_value=httpx.Response(200, text=data))
    # Responses for plugin folders
    respx.get("http://perdu.com/plugins/oembed/").mock(return_value=httpx.Response(403))
    respx.get("http://perdu.com/plugins-dist/nuage/").mock(return_value=httpx.Response(403))


    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(403))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 3

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20, "cms": "spip"}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "SPIP", "versions": [], "categories": ["CMS SPIP"], "groups": ["Content"]}'
        )



@pytest.mark.asyncio
@respx.mock
async def test_spip_no_plugins_403():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/spip/")
    spip_edited = "CHANGELOG_edited.txt"

    with open(path_join(test_directory, spip_edited), errors="ignore") as spip:
        data = spip.read()

    # Response to tell that SPIP is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="This website use SPIP, your_spip_attribute = ''",
            headers={"composed-by": "SPIP"})
    )

    # Response for edited changelog.txt
    respx.get("http://perdu.com/CHANGELOG.txt").mock(return_value=httpx.Response(200, text=data))
    # Responses for plugin folders
    respx.get("http://perdu.com/plugins/oembed/").mock(return_value=httpx.Response(403))
    respx.get("http://perdu.com/plugins-dist/nuage/").mock(return_value=httpx.Response(403))


    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(403))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 3

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20, "cms": "spip"}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "SPIP", "versions": [], "categories": ["CMS SPIP"], "groups": ["Content"]}'
        )



@pytest.mark.asyncio
@respx.mock
async def test_typo3_version_not_detected():
    mocked_content = (
        '<html><head><link rel="stylesheet" type="text/css" '
        'This website use TYPO3, <img src=typo3conf />">'
        '</head><body>This website uses Typo3</body></html>'
    )

    # Response to tell that TYPO3 is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="This website use TYPO3, <img src=typo3conf />")
    )

    respx.get("http://perdu.com/typo3/").mock(
        return_value=httpx.Response(
            200,
            content="This website use TYPO3, Login page <img src=typo3conf />")
    )

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    # Mocking browser behavior - mock playwright_get_content to return a mock page content
    with patch("wapitiCore.attack.cms.mod_typo3_enum.playwright_get_content", return_value=mocked_content):

        # Mock the persister
        persister = AsyncMock()
        persister.get_root_url.return_value = "http://perdu.com/"

        # Setup request and crawler configuration
        request = Request("http://perdu.com/")
        request.path_id = 1

        crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
        async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
            options = {"timeout": 10, "level": 2, "tasks": 20, "cms": "typo3"}
            module = ModuleCms(crawler, persister, options, crawler_configuration)

            # Run the attack
            await module.attack(request)

            # Assertions for added payloads
            assert persister.add_payload.call_count == 1
            assert persister.add_payload.call_args_list[0][1]["info"] == (
                '{"name": "TYPO3", "versions": [], "categories": ["CMS TYPO3"], "groups": ["Content"]}'
            )

@pytest.mark.asyncio
@respx.mock
async def test_typo3_via_script_tag():
    mocked_content = (
        '<html><head><script src="/typo3conf/ext/some-extension/script.js"></script>'
        '</head><body>This website uses Typo3<script src="typo3temp/assets/js/main.js"></script></body></html>'
    )

    # Response to tell that TYPO3 is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content='This website use TYPO3, <script src="typo3temp/assets/js/main.js"></script>')
    )

    respx.get("http://perdu.com/typo3/").mock(
        return_value=httpx.Response(
            200,
            content='This website use TYPO3, Login page <script src="typo3temp/assets/js/main.js"></script>')
    )

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    # Mocking browser behavior - mock playwright_get_content to return a mock page content
    with patch("wapitiCore.attack.cms.mod_typo3_enum.playwright_get_content", return_value=mocked_content):

        # Mock the persister
        persister = AsyncMock()
        persister.get_root_url.return_value = "http://perdu.com/"

        # Setup request and crawler configuration
        request = Request("http://perdu.com/")
        request.path_id = 1

        crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
        async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
            options = {"timeout": 10, "level": 2, "tasks": 20, "cms": "typo3"}
            module = ModuleCms(crawler, persister, options, crawler_configuration)

            # Run the attack
            await module.attack(request)

            # Assertions for added payloads
            assert persister.add_payload.call_count == 1
            assert persister.add_payload.call_args_list[0][1]["info"] == (
                '{"name": "TYPO3", "versions": [], "categories": ["CMS TYPO3"], "groups": ["Content"]}'
            )

@pytest.mark.asyncio
@respx.mock
async def test_typo3_via_image_svg():
    mocked_content = (
        '<html><head><script src="/typo3conf/ext/some-extension/script.js"></script>'
        '</head><body>This website uses Typo3<script src="typo3temp/assets/js/main.js"></script></body></html>'
    )

    # Response to tell that TYPO3 is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content='Hello World')
    )

    respx.get("http://perdu.com/typo3/sysext/core/Resources/Public/Images/typo3_orange.svg").mock(
        return_value=httpx.Response(
            200,
            headers={"Content-Type": "image/svg+xml"},
            content='image svg typo3')
    )

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    # Mocking browser behavior - mock playwright_get_content to return a mock page content
    with patch("wapitiCore.attack.cms.mod_typo3_enum.playwright_get_content", return_value=mocked_content):

        # Mock the persister
        persister = AsyncMock()
        persister.get_root_url.return_value = "http://perdu.com/"

        # Setup request and crawler configuration
        request = Request("http://perdu.com/")
        request.path_id = 1

        crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
        async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
            options = {"timeout": 10, "level": 2, "tasks": 20, "cms": "typo3"}
            module = ModuleCms(crawler, persister, options, crawler_configuration)

            # Run the attack
            await module.attack(request)

            # Assertions for added payloads
            assert persister.add_payload.call_count == 1
            assert persister.add_payload.call_args_list[0][1]["info"] == (
                '{"name": "TYPO3", "versions": [], "categories": ["CMS TYPO3"], "groups": ["Content"]}'
            )



@pytest.mark.asyncio
@respx.mock
async def test_typo3_version_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/typo3/")
    typo_shims_file = "es-module-shims.js"

    with open(path_join(test_directory, typo_shims_file), errors="ignore") as typo:
        data = typo.read()

    # Response to tell that TYPO3 is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="This website use TYPO3, <img src=typo3conf />")
    )

    respx.get("http://perdu.com/typo3/").mock(
        return_value=httpx.Response(
            200,
            content="This website use TYPO3, Login page <img src=typo3conf />")
    )

    # Response for es-module-shims.js
    respx.get("http://perdu.com/typo3/sysext/core/Resources/Public/JavaScript/Contrib/es-module-shims.js").mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20, "cms": "typo3"}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 2
        assert persister.add_payload.call_args_list[0][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[0][1]["category"] == "Fingerprint web application framework"
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "TYPO3", "versions": ["v12.4.0"], "categories": ["CMS TYPO3"], "groups": ["Content"]}'
        )
        assert persister.add_payload.call_args_list[1][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[1][1]["category"] == "Fingerprint web technology"
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "TYPO3", "versions": ["v12.4.0"], "categories": ["CMS TYPO3"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_typo3_multi_versions_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/typo3/")
    require_file = "require.js"

    with open(path_join(test_directory, require_file), errors="ignore") as requirejs:
        data = requirejs.read()

    # Response to tell that TYPO3 is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="This website use TYPO3, <img src=typo3conf />")
    )

    respx.get("http://perdu.com/typo3/").mock(
        return_value=httpx.Response(
            200,
            content="This website use TYPO3, Login page <img src=typo3conf />")
    )

    # Response for require.js
    respx.get("http://perdu.com/typo3/sysext/core/Resources/Public/JavaScript/Contrib/require.js")\
        .mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 2
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "TYPO3", "versions": ["v11.5.39", "v11.5.40", "v11.5.41"], "categories": ["CMS TYPO3"], "groups": ["Content"]}'
        )
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "TYPO3", "versions": ["v11.5.39", "v11.5.40", "v11.5.41"], "categories": ["CMS TYPO3"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_typo3_extension():
    mocked_content = (
        '<html><head><link rel="stylesheet" type="text/css" '
        'This website use TYPO3, <img src=typo3conf />">'
        '</head><body>This website uses Typo3</body></html>'
    )

    # Response to tell that TYPO3 is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="This website use TYPO3, <img src=typo3conf />")
    )

    respx.get("http://perdu.com/typo3/").mock(
        return_value=httpx.Response(
            200,
            content="This website use TYPO3, Login page <img src=typo3conf />")
    )

    respx.get("http://perdu.com/typo3conf/ext/blog/").mock(
        return_value=httpx.Response(
            403,
            content="Forbidden")
    )

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    # Mocking browser behavior - mock playwright_get_content to return a mock page content
    with patch("wapitiCore.attack.cms.mod_typo3_enum.playwright_get_content", return_value=mocked_content):

        # Mock the persister
        persister = AsyncMock()
        persister.get_root_url.return_value = "http://perdu.com/"

        # Setup request and crawler configuration
        request = Request("http://perdu.com/")
        request.path_id = 1

        crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
        async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
            options = {"timeout": 10, "level": 2, "tasks": 20, "cms": "typo3"}
            module = ModuleCms(crawler, persister, options, crawler_configuration)

            # Run the attack
            await module.attack(request)

            # Assertions for added payloads
            assert persister.add_payload.call_count == 2
            assert persister.add_payload.call_args_list[0][1]["info"] == (
                '{"name": "TYPO3", "versions": [], "categories": ["CMS TYPO3"], "groups": ["Content"]}'
            )
            assert persister.add_payload.call_args_list[1][1]["info"] == (
                '{"name": "blog", "versions": [], "categories": ["TYPO3 extension"], "groups": ["Add-ons"]}'
            )


@pytest.mark.asyncio
@respx.mock
async def test_typo3_no_extension_403():
    mocked_content = (
        '<html><head><meta name="generator" content="TYPO3 CMS 11.5.2">'
        '<link rel="stylesheet" type="text/css" This website use TYPO3, <img src=typo3conf />">'
        '</head><body>Hello World</body></html>'
    )

    # Response to tell that TYPO3 is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content='<html><head><meta name="generator" content="TYPO3 CMS 11.5.2"></head><body>hello</body></html>')
    )

    respx.get("http://perdu.com/typo3/").mock(
        return_value=httpx.Response(
            200,
            content='<html><head><meta name="generator" content="TYPO3 CMS 11.5.2"></head><body>hello</body></html>')
    )

    respx.get("http://perdu.com/typo3conf/ext/blog/").mock(
        return_value=httpx.Response(
            403,
            content="Forbidden")
    )

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(403))

    # Mocking browser behavior - mock playwright_get_content to return a mock page content
    with patch("wapitiCore.attack.cms.mod_typo3_enum.playwright_get_content", return_value=mocked_content):

        # Mock the persister
        persister = AsyncMock()
        persister.get_root_url.return_value = "http://perdu.com/"

        # Setup request and crawler configuration
        request = Request("http://perdu.com/")
        request.path_id = 1

        crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
        async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
            options = {"timeout": 10, "level": 2, "tasks": 20, "cms": "typo3"}
            module = ModuleCms(crawler, persister, options, crawler_configuration)

            # Run the attack
            await module.attack(request)

            # Assertions for added payloads
            assert persister.add_payload.call_count == 1

@pytest.mark.asyncio
@respx.mock
async def test_fetch_source_files_typo3():
    base_url = "http://perdu.com/"

    # Mocked HTML content with JS and CSS files
    mocked_content = '''
    <html>
        <head>
            <script src="/static/js/script1.js"></script>
            <script src="http://perdu.com/static/js/script2.js"></script>
            <link rel="stylesheet" href="/static/css/style1.css">
            <link rel="stylesheet" href="http://perdu.com/static/css/style2.css">
        </head>
        <body>Test page</body>
    </html>
    '''

    # Mocking browser behavior - mock playwright_get_content to return a mock page content
    with patch("wapitiCore.attack.cms.mod_typo3_enum.playwright_get_content", return_value=mocked_content):
        crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
        # Run the function
        result = await fetch_source_files_typo3(base_url, crawler_configuration)

        # Expected extracted files
        expected_files = {
            "/static/js/script1.js",
            "http://perdu.com/static/js/script2.js",
            "/static/css/style1.css",
            "http://perdu.com/static/css/style2.css",
        }

        # Validate output
        assert result == expected_files


@pytest.mark.asyncio
@respx.mock
async def test_wp_version_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/wp/")
    wp_file = "media-views-rtl.min.css"

    with open(path_join(test_directory, wp_file), errors="ignore") as wp_style:
        data = wp_style.read()

    # Response to tell that WordPress is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Ce site utilise Wordpress .... /wp-content</h2> \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>")
    )

    # Response for media-views-rtl.min.css
    respx.get("http://perdu.com/wp-includes/css/media-views-rtl.min.css").mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 2
        assert persister.add_payload.call_args_list[0][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[0][1]["category"] == "Fingerprint web application framework"
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "WordPress", "versions": ["3.8"], "categories": ["CMS WordPress"], "groups": ["Content"]}'
        )
        assert persister.add_payload.call_args_list[1][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[1][1]["category"] == "Fingerprint web technology"
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "WordPress", "versions": ["3.8"], "categories": ["CMS WordPress"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_wp_multi_versions_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/wp/")
    style_file = "style.min.css"

    with open(path_join(test_directory, style_file), errors="ignore") as wp_style:
        data = wp_style.read()

    # Response to tell that WordPress is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Ce site utilise Wordpress .... /wp-content</h2> \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>")
    )

    # Response for style.min.css
    respx.get("http://perdu.com/wp-includes/css/dist/block-library/style.min.css")\
        .mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 2
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "WordPress", "versions": ["5.0", "5.0.1", "5.0.2"], "categories": ["CMS WordPress"], "groups": ["Content"]}'
        )
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "WordPress", "versions": ["5.0", "5.0.1", "5.0.2"], "categories": ["CMS WordPress"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_wp_no_version_detected():

    base_dir = os.path.dirname(sys.modules["wapitiCore"].__file__)
    test_directory = os.path.join(base_dir, "..", "tests/data/wp/")
    wp_file = "edited_media-views-rtl.min.css"

    with open(path_join(test_directory, wp_file), errors="ignore") as wp_style:
        data = wp_style.read()

    # Response to tell that WordPress is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Ce site utilise Wordpress .... /wp-content</h2> \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>")
    )

    # Response for media-views-rtl.min.css
    respx.get("http://perdu.com/wp-includes/css/media-views-rtl.min.css").mock(return_value=httpx.Response(200, text=data))

    respx.get(url__regex=r"http://perdu.com/.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com/")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 1
        assert persister.add_payload.call_args_list[0][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "WordPress", "versions": [], "categories": ["CMS WordPress"], "groups": ["Content"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_wp_false_positive_403():

    # Response to tell that WordPress is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Ce site utilise Wordpress .... /wp-content</h2> \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>")
    )

    # Response for versioned plugin
    respx.get("http://perdu.com/wp-content/plugins/bbpress/readme.txt").mock(
        return_value=httpx.Response(
            200,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Pas de panique, on va wordpress vous aider</h2> \
            Wordpress wordpress wp-content\
            Stable tag: 2.6.6 \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    # Response for plugin detected without version (403 forbiden response)
    respx.get("http://perdu.com/wp-content/plugins/wp-reset/readme.txt").mock(
        return_value=httpx.Response(403)
    )

    # Response for bad format readme.txt of plugin
    respx.get("http://perdu.com/wp-content/plugins/unyson/readme.txt").mock(
        return_value=httpx.Response(
            200,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Pas de panique, on va wordpress vous aider</h2> \
            Wordpress wordpress wp-content\
            Version Tested : 4.5 \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    # Response for versioned theme
    respx.get("http://perdu.com/wp-content/themes/twentynineteen/readme.txt").mock(
        return_value=httpx.Response(
            200,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
               <h2>Pas de panique, on va wordpress vous aider</h2> \
               Wordpress wordpress WordPress\
               Stable tag: 1.9 \
               <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    # Response for theme detected without version (403 forbidden response)
    respx.get("http://perdu.com/wp-content/themes/seedlet/readme.txt").mock(
        return_value=httpx.Response(403)
    )

    # Response for bad format readme.txt of theme
    respx.get("http://perdu.com/wp-content/themes/customify/readme.txt").mock(
        return_value=httpx.Response(
            200,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Pas de panique, on va wordpress vous aider</h2> \
            Wordpress wp-content WordPress\
            Version Tested : 3.2 \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    respx.get(url__regex=r"http://perdu.com/wp-content/plugins/.*?/readme.txt").mock(return_value=httpx.Response(403))
    respx.get(url__regex=r"http://perdu.com/wp-content/themes/.*?/readme.txt").mock(return_value=httpx.Response(403))
    respx.get(url__regex=r"http://perdu.com.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 5

        assert persister.add_payload.call_args_list[0][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[0][1]["category"] == "Fingerprint web technology"
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "WordPress", "versions": [], "categories": ["CMS WordPress"], "groups": ["Content"]}'
        )

        assert persister.add_payload.call_args_list[1][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[1][1]["category"] == "Fingerprint web technology"
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "bbpress", "versions": ["2.6.6"], "categories": ["WordPress plugins"], "groups": ["Add-ons"]}'
        )
        assert persister.add_payload.call_args_list[2][1]["info"] == (
            '{"name": "unyson", "versions": [], "categories": ["WordPress plugins"], "groups": ["Add-ons"]}'
        )
        assert persister.add_payload.call_args_list[3][1]["info"] == (
            '{"name": "twentynineteen", "versions": ["1.9"], "categories": ["WordPress themes"], "groups": ["Add-ons"]}'
        )
        assert persister.add_payload.call_args_list[4][1]["info"] == (
            '{"name": "customify", "versions": [], "categories": ["WordPress themes"], "groups": ["Add-ons"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_wp_false_positive_success():

    # Response to tell that WordPress is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Ce site utilise Wordpress .... /wp-content</h2> \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>")
    )

    # Response for versioned plugin
    respx.get("http://perdu.com/wp-content/plugins/bbpress/readme.txt").mock(
        return_value=httpx.Response(
            200,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Pas de panique, on va wordpress vous aider</h2> \
            Wordpress wordpress wp-content\
            Stable tag: 2.6.6 \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    # Response for plugin detected without version (403 forbiden response)
    respx.get("http://perdu.com/wp-content/plugins/wp-reset/readme.txt").mock(
        return_value=httpx.Response(403)
    )

    # Response for bad format readme.txt of plugin
    respx.get("http://perdu.com/wp-content/plugins/unyson/readme.txt").mock(
        return_value=httpx.Response(
            200,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Pas de panique, on va wordpress vous aider</h2> \
            Wordpress wordpress wp-content\
            Version Tested : 4.5 \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    # Response for versioned theme
    respx.get("http://perdu.com/wp-content/themes/twentynineteen/readme.txt").mock(
        return_value=httpx.Response(
            200,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
               <h2>Pas de panique, on va wordpress vous aider</h2> \
               Wordpress wordpress WordPress\
               Stable tag: 1.9 \
               <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    # Response for theme detected without version (403 forbidden response)
    respx.get("http://perdu.com/wp-content/themes/seedlet/readme.txt").mock(
        return_value=httpx.Response(403)
    )

    # Response for bad format readme.txt of theme
    respx.get("http://perdu.com/wp-content/themes/customify/readme.txt").mock(
        return_value=httpx.Response(
            200,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Pas de panique, on va wordpress vous aider</h2> \
            Wordpress wp-content WordPress\
            Version Tested : 3.2 \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    respx.get(url__regex=r"http://perdu.com/wp-content/plugins/.*?/readme.txt").mock(return_value=httpx.Response(200))
    respx.get(url__regex=r"http://perdu.com/wp-content/themes/.*?/readme.txt").mock(return_value=httpx.Response(200))
    respx.get(url__regex=r"http://perdu.com.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count == 5

        assert persister.add_payload.call_args_list[0][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[0][1]["category"] == "Fingerprint web technology"
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "WordPress", "versions": [], "categories": ["CMS WordPress"], "groups": ["Content"]}'
        )

        assert persister.add_payload.call_args_list[1][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[1][1]["category"] == "Fingerprint web technology"
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "bbpress", "versions": ["2.6.6"], "categories": ["WordPress plugins"], "groups": ["Add-ons"]}'
        )
        assert persister.add_payload.call_args_list[2][1]["info"] == (
            '{"name": "wp-reset", "versions": [], "categories": ["WordPress plugins"], "groups": ["Add-ons"]}'
        )
        assert persister.add_payload.call_args_list[3][1]["info"] == (
            '{"name": "twentynineteen", "versions": ["1.9"], "categories": ["WordPress themes"], "groups": ["Add-ons"]}'
        )
        assert persister.add_payload.call_args_list[4][1]["info"] == (
            '{"name": "seedlet", "versions": [], "categories": ["WordPress themes"], "groups": ["Add-ons"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_wp_plugin():

    # Response to tell that WordPress is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Ce site utilise Wordpress .... /wp-content</h2> \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>")
    )

    # Response for versioned plugin
    respx.get("http://perdu.com/wp-content/plugins/bbpress/readme.txt").mock(
        return_value=httpx.Response(
            200,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Pas de panique, on va wordpress vous aider</h2> \
            Wordpress wordpress wp-content\
            Stable tag: 2.6.6 \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    # Response for plugin detected without version (403 forbiden response)
    respx.get("http://perdu.com/wp-content/plugins/wp-reset/readme.txt").mock(
        return_value=httpx.Response(
            403,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Pas de panique, on va wordpress vous aider</h2> \
            Wordpress wordpress wp-content\
            Stable tag: 9.5.1 \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    # Response for bad format readme.txt of plugin
    respx.get("http://perdu.com/wp-content/plugins/unyson/readme.txt").mock(
        return_value=httpx.Response(
            200,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Pas de panique, on va wordpress vous aider</h2> \
            Wordpress wordpress wp-content\
            Version Tested : 4.5 \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    respx.get(url__regex=r"http://perdu.com/wp-content/plugins/.*?/readme.txt").mock(return_value=httpx.Response(404))
    respx.get(url__regex=r"http://perdu.com/wp-content/themes/.*?/readme.txt").mock(return_value=httpx.Response(404))
    respx.get(url__regex=r"http://perdu.com.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count

        assert persister.add_payload.call_args_list[0][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[0][1]["category"] == "Fingerprint web technology"
        assert persister.add_payload.call_args_list[0][1]["info"] == (
            '{"name": "WordPress", "versions": [], "categories": ["CMS WordPress"], "groups": ["Content"]}'
        )

        assert persister.add_payload.call_args_list[1][1]["module"] == "cms"
        assert persister.add_payload.call_args_list[1][1]["category"] == "Fingerprint web technology"
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "bbpress", "versions": ["2.6.6"], "categories": ["WordPress plugins"], "groups": ["Add-ons"]}'
        )
        assert persister.add_payload.call_args_list[2][1]["info"] == (
            '{"name": "wp-reset", "versions": [], "categories": ["WordPress plugins"], "groups": ["Add-ons"]}'
        )
        assert persister.add_payload.call_args_list[3][1]["info"] == (
            '{"name": "unyson", "versions": [], "categories": ["WordPress plugins"], "groups": ["Add-ons"]}'
        )


@pytest.mark.asyncio
@respx.mock
async def test_wp_theme():

    # Response to tell that WordPress is used
    respx.get("http://perdu.com/").mock(
        return_value=httpx.Response(
            200,
            content="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Ce site utilise Wordpress .... /wp-content</h2> \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>")
    )

    # Response for versioned theme
    respx.get("http://perdu.com/wp-content/themes/twentynineteen/readme.txt").mock(
        return_value=httpx.Response(
            200,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
               <h2>Pas de panique, on va wordpress vous aider</h2> \
               Wordpress wordpress WordPress\
               Stable tag: 1.9 \
               <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    # Response for theme detected without version (403 forbidden response)
    respx.get("http://perdu.com/wp-content/themes/seedlet/readme.txt").mock(
        return_value=httpx.Response(
            403,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Pas de panique, on va wordpress vous aider</h2> \
            Wordpress wp-content WordPress\
            Stable tag: 5.4 \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    # Response for bad format readme.txt of theme
    respx.get("http://perdu.com/wp-content/themes/customify/readme.txt").mock(
        return_value=httpx.Response(
            200,
            text="<html><head><title>Vous Etes Perdu ?</title></head><body><h1>Perdu sur l'Internet ?</h1> \
            <h2>Pas de panique, on va wordpress vous aider</h2> \
            Wordpress wp-content WordPress\
            Version Tested : 3.2 \
            <strong><pre>    * <----- vous &ecirc;tes ici</pre></strong></body></html>"
        )
    )

    respx.get(url__regex=r"http://perdu.com/wp-content/plugins/.*?/readme.txt").mock(return_value=httpx.Response(404))
    respx.get(url__regex=r"http://perdu.com/wp-content/themes/.*?/readme.txt").mock(return_value=httpx.Response(404))
    respx.get(url__regex=r"http://perdu.com.*?").mock(return_value=httpx.Response(404))

    persister = AsyncMock()

    request = Request("http://perdu.com")
    request.path_id = 1

    crawler_configuration = CrawlerConfiguration(Request("http://perdu.com/"))
    async with AsyncCrawler.with_configuration(crawler_configuration) as crawler:
        options = {"timeout": 10, "level": 2, "tasks": 20}

        module = ModuleCms(crawler, persister, options, crawler_configuration)

        await module.attack(request)

        assert persister.add_payload.call_count
        assert persister.add_payload.call_args_list[1][1]["info"] == (
            '{"name": "twentynineteen", "versions": ["1.9"], "categories": ["WordPress themes"], "groups": ["Add-ons"]}'
        )
        assert persister.add_payload.call_args_list[2][1]["info"] == (
            '{"name": "seedlet", "versions": [], "categories": ["WordPress themes"], "groups": ["Add-ons"]}'
        )
        assert persister.add_payload.call_args_list[3][1]["info"] == (
            '{"name": "customify", "versions": [], "categories": ["WordPress themes"], "groups": ["Add-ons"]}'
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the Wapiti project (https://wapiti-scanner.github.io)
# Copyright (C) 2025 Cyberwatch
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
import json
import re
from typing import Optional
from urllib.parse import urljoin
from string import ascii_lowercase
import random

from wapitiCore.parsers.html_parser import Html
from httpx import RequestError

from wapitiCore.attack.network_devices.network_device_common import NetworkDeviceCommon, MSG_TECHNO_VERSIONED
from wapitiCore.definitions.cve_2024_55591 import FortiAuthenticationBypass
from wapitiCore.net import Request
from wapitiCore.net.response import Response
from wapitiCore.definitions.fingerprint import SoftwareNameDisclosureFinding
from wapitiCore.main.log import log_blue, logging, log_verbose, log_red

MSG_NO_FORTI = "No Forti Product Detected"
MSG_CVE_2024_55591 = "[!] {0} is vulnerable to CVE-2024-55591 !"


class ModuleForti(NetworkDeviceCommon):
    """Detect Forti."""
    device_name = "Fortinet"
    version = []
    fortinet_pattern = re.compile(r'Forti\w+')

    async def check_forti(self, url):
        fortivpn_list = ['remote/fgt_lang?lang=en',
                         'remote/fgt_lang?lang=fr']
        fortiweb_list = ['fgt_lang.js?paths=lang/en:com_info',
                         'fgt_lang.js?paths=lang/fr:com_info']
        check_list = fortivpn_list + fortiweb_list + [
            'logindisclaimer',
            'remote/login?lang=en',
            'remote/login?lang=fr',
            'fpc/app/login',
            'login/?next=/',
            'login?redir='
        ]

        # Check FortiMail
        url_fortimail = urljoin(url, "admin/")
        request = Request(url_fortimail, 'GET')
        try:
            response: Response = await self.crawler.async_send(request, follow_redirects=False)
            log_verbose(f"[¨] {request}")
        except RequestError:
            self.network_errors += 1
            raise
        if response.is_success:
            page = Html(response.content, url_fortimail)
            for tag in page.soup.find_all(True):
                if tag.string:
                    match = self.fortinet_pattern.search(tag.text)
                    if match:
                        # Extract the matched product name
                        self.device_name = match.group()
                        return True

        # Check FortiManager and FortiAnalyzer
        url_fortimanager = urljoin(url, "p/login/")
        request = Request(url_fortimanager, 'GET')
        try:
            response: Response = await self.crawler.async_send(request, follow_redirects=False)
            log_verbose(f"[¨] {request}")
        except RequestError:
            self.network_errors += 1
            raise
        if response.is_success:
            page = Html(response.content, url_fortimanager)
            sign_in_header_div = page.soup.find('div', class_='sign-in-header')

            for device_name in ["FortiManager", "FortiAnalyzer"]:
                if device_name in page.title:
                    self.device_name = device_name
                    return True
                # if custom title without Forti*, we check for specific div
                if sign_in_header_div and device_name in sign_in_header_div.text:
                    self.device_name = device_name
                    return True

        # Check for other Fortinet products
        for item in check_list:
            full_url = urljoin(url, item)
            request = Request(full_url, 'GET')
            try:
                response: Response = await self.crawler.async_send(request, follow_redirects=False)
                log_verbose(f"[¨] {request}")
            except RequestError:
                self.network_errors += 1
                continue

            if response.is_success:
                if "content-type" in response.headers and \
                        "javascript" in response.headers["content-type"]:
                    if item in fortivpn_list:
                        self.device_name = "Fortinet SSL-VPN"
                        return True
                    if item in fortiweb_list:
                        self.device_name = "FortiWeb"
                        return True
                return await self.detect_forti_product(full_url)
        return False

    async def detect_forti_product(self, url):
        request = Request(url, 'GET')
        try:
            # Send an HTTP GET request to the URL
            response: Response = await self.crawler.async_send(request, follow_redirects=True)
            log_verbose(f"[¨] {request}")
        except RequestError:
            self.network_errors += 1
            raise
        page = Html(response.content, url)

        match = self.fortinet_pattern.search(page.title)
        if match:
            self.device_name = match.group()
            return True
        if 'class="main-app"' in response.content:
            return True

        return False

    async def detect_cve_2024_55591(self, base_url):
        """
        Detects if the Fortinet FortiOS or FortiProxy is vulnerable to CVE-2024-55591.
        This function was adapted from the WatchTowrLabs POC project of detecting the CVE-2024-55591 vulnerability
        Source: https://github.com/watchtowrlabs/fortios-auth-bypass-check-CVE-2024-55591
        """

        #Send a request to a specific endpoint
        login_endpoint = "login?redir=/ng"
        login_url = urljoin(base_url, login_endpoint)
        login_request = Request(login_url, 'GET')
        try:
            # Send the WebSocket-like request
            login_response: Response = await self.crawler.async_send(login_request, follow_redirects=False)
        except RequestError:
            self.network_errors += 1
            raise

        # Generate a random endpoint suffix
        random_suffix = ''.join(random.choices(ascii_lowercase, k=8))
        test_uri = f"wapiti3-{random_suffix}"
        test_url = urljoin(base_url, test_uri)

        # WebSocket handshake headers
        test_headers = {
            "Sec-WebSocket-Version": "13",
            "Sec-WebSocket-Key": "thFz/fKwzu5wDEy0XO3fcw==",
            "Connection": "keep-alive, Upgrade",
            "Upgrade": "websocket"
        }
        request = Request(test_url, 'GET')
        try:
            # Send the WebSocket-like request
            test_response: Response = await self.crawler.async_send(request, follow_redirects=False, headers=test_headers)
        except RequestError:
            self.network_errors += 1
            raise

        # Analyze the response status code
        login_request_status_check = login_response.status == 200
        test_request_status_check = test_response.status == 101

        # Check body content validation

        main_app_check = '<html class="main-app">' in login_response.content
        f_icon_warning_check = '<f-icon class="fa-warning' in login_response.content
        f_icon_closing_check = '</f-icon>' in login_response.content

        body_checks = main_app_check and f_icon_warning_check and f_icon_closing_check

        # Check for APSCOOKIE header
        header_marker_check = any('APSCOOKIE_' in str(header) for header in login_response.headers.values())

        # Check connection upgrade for second response
        connection_upgrade_check = 'Upgrade' in test_response.headers.get('Connection', '')

        return all([
            login_request_status_check,
            test_request_status_check,
            body_checks,
            header_marker_check,
            connection_upgrade_check
        ])

    async def attack(self, request: Request, response: Optional[Response] = None):
        self.finished = True
        request_to_root = Request(request.url)

        try:
            if await self.check_forti(request_to_root.url):

                forti_detected = {
                    "name": self.device_name,
                    "versions": self.version,
                    "categories": ["Network Equipment"],
                    "groups": ["Content"]
                }
                log_blue(
                    MSG_TECHNO_VERSIONED,
                    self.device_name,
                    self.version
                )

                await self.add_info(
                    finding_class=SoftwareNameDisclosureFinding,
                    request=request_to_root,
                    info=json.dumps(forti_detected),
                )

                if await self.detect_cve_2024_55591(request.url):
                    log_red(
                        MSG_CVE_2024_55591,
                        request_to_root.url
                    )
                    await self.add_critical(
                        finding_class=FortiAuthenticationBypass,
                        request=request,
                        info=f"URL {request.url} seems vulnerable to CVE-2024-55591",
                        response=response
                    )
                else:
                    log_blue("Not vulnerable to CVE-2024-55591")

            else:
                log_blue(MSG_NO_FORTI)
        except RequestError as req_error:
            self.network_errors += 1
            logging.error(f"Request Error occurred: {req_error}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the Wapiti project (https://wapiti-scanner.github.io)
# Copyright (C) 2023-2025 Cyberwatch
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
import json
import re
from os.path import join as path_join
from typing import Optional, List, Set, Dict

from urllib.parse import urljoin
from httpx import RequestError
from asyncio import TimeoutError
from playwright.async_api import async_playwright, Page, Browser, Error as PlaywrightError

from wapitiCore.parsers.html_parser import Html
from wapitiCore.net import Request
from wapitiCore.attack.cms.cms_common import CommonCMS, MSG_TECHNO_VERSIONED
from wapitiCore.net.classes import CrawlerConfiguration
from wapitiCore.net.response import Response
from wapitiCore.definitions.fingerprint_webapp import SoftwareVersionDisclosureFinding
from wapitiCore.definitions.fingerprint import SoftwareNameDisclosureFinding
from wapitiCore.main.log import logging, log_blue

MSG_NO_TYPO3 = "No TYPO3 Detected"


async def get_common_versions(json_data: Dict, list_hashes: List[str]) -> list[str]:
    """
    Given a list of hashes and a JSON file, this function returns the set of
    versions that are common across all occurrences of the provided hashes in the JSON."""

    # To store version sets for each hash found across the JSON
    hash_versions = {h: set() for h in list_hashes}
    # Iterate over each file path in the JSON data.
    for file_path, hashes in json_data.items():
        # hashes is a dict mapping a hash to a list of versions.
        for h in list_hashes:
            if h in hashes:
                hash_versions[h].update(hashes[h])

    # If any hash did not appear in the JSON, its version set remains empty.
    # In that case, we should remove all the empty entries from the list.

    hash_versions = {key: versions for key, versions in hash_versions.items() if len(versions) > 0}

    # Compute the intersection across all version sets.
    common_versions = set.intersection(*hash_versions.values()) if hash_versions else []
    return list(common_versions)


async def playwright_get_content(url: str, page: Page, browser: Browser, crawler_configuration: CrawlerConfiguration) -> str:
    html = ""
    try:
        await page.goto(
            url,
            timeout = crawler_configuration.timeout * 1000,
            wait_until="load"
        )

        html = await page.content()
        return html
    except PlaywrightError as exception:
        logging.exception(exception)
        return ""
    finally:
        await browser.close()
        


async def fetch_source_files(url: str, crawler_configuration: CrawlerConfiguration) -> set:
    my_files_list = set()
    url =  urljoin(url, "typo3/")
    proxy_settings = None
    if crawler_configuration.proxy:
        proxy_settings = {"server": crawler_configuration.proxy}

    try:
        # Initialize a headless browser session to extract source files
        async with async_playwright() as pw:
            browser = await pw.firefox.launch(
                headless=True,
                proxy=proxy_settings,
                firefox_user_prefs={
                    "network.proxy.allow_hijacking_localhost": True,
                    "devtools.jsonview.enabled": False,
                }
            )
            context = await browser.new_context(
                ignore_https_errors=True, 
                user_agent=crawler_configuration.user_agent
            )
            page = await context.new_page()

            html = await playwright_get_content(url, page, browser, crawler_configuration)
            parsed = Html(html, url)

            # Collect JS and CSS file links
            js_files = {script['src'] for script in parsed.soup.find_all('script', src=True)}
            css_files = {link['href'] for link in parsed.soup.find_all('link', rel="stylesheet", href=True)}

            my_files_list.update(js_files)
            my_files_list.update(css_files)

    except (PlaywrightError, FileNotFoundError, TimeoutError) as exception:
            # Playwright browser may be missing, etc
            logging.exception(exception)
            logging.warning(
                "Could not run headless browser. "
                "Make sure playwright is installed (`pip install playwright`) "
                "and browsers are installed (`playwright install`)."
            )

    except Exception as e:
        logging.error("An error occurred while fetching JS files: %s", e)

    return my_files_list


class ModuleTYPO3Enum(CommonCMS):
    """Detect TYPO3 version."""
    PAYLOADS_HASH = "typo3_hash_files.json"
    PAYLOADS_FILE_EXTENSIONS = "typo3_extensions.txt"
    versions = []
    extensions_list = []

    async def check_typo3_extensions(self, url):
        """
        Check if specific TYPO3 extensions are installed on the given URL.
        """
        installed_extensions = []

        no_ext_url = urljoin(url, "typo3conf/ext/non_existing_ext/")
        no_ext_request = Request(no_ext_url, 'GET')

        try:
            no_ext_response: Response = await self.crawler.async_send(no_ext_request, follow_redirects=True)
            if no_ext_response.status == 403:
                return []
        except RequestError:
            self.network_errors += 1
            return []

        extensions_path = path_join(self.DATA_DIR, self.PAYLOADS_FILE_EXTENSIONS)
        try:
            with open(extensions_path, errors="ignore", encoding="utf-8") as ext_list:
                for extension in ext_list:
                    extension = extension.strip()
                    ext_url = urljoin(url, f"typo3conf/ext/{extension}/")
                    request = Request(ext_url, 'GET')
                    try:
                        response: Response = await self.crawler.async_send(request, follow_redirects=True)
                    except RequestError:
                        self.network_errors += 1
                        continue

                    if response.status == 403:
                        installed_extensions.append(extension)

        except FileNotFoundError:
            logging.error("TYPO3 extensions file not found: %s", extensions_path)
            return []

        return installed_extensions


    async def check_typo3(self, url):

        request = Request(f'{url}', 'GET')
        try:
            response: Response = await self.crawler.async_send(request, follow_redirects=True)
        except RequestError:
            self.network_errors += 1
        else:
            page = Html(response.content, url)

            # Check meta tag for generator
            meta_generator = page.soup.find('meta', attrs={'name': 'generator'})
            if meta_generator and re.search(r'TYPO3\s+(?:CMS\s+)?(?:[\d.]+)?(?:\s+CMS)?',
                                            meta_generator.get('content', ''), re.I):
                return True

            # Check for TYPO3-specific links and images
            typo3_patterns = ['typo3conf', 'typo3temp']
            for tag in page.soup.find_all(['link', 'img']):
                for attr in ['href', 'src']:
                    if tag.has_attr(attr) and any(pattern in tag[attr] for pattern in typo3_patterns):
                        return True

            # Check script sources
            for script in page.soup.find_all('script', src=True):
                if re.search(r'^/?typo3(?:conf|temp)/', script['src']):
                    return True

            # Check for a known TYPO3 image probe
            typo3_probe_url = url.rstrip('/') + "/typo3/sysext/core/Resources/Public/Images/typo3_orange.svg"
            request_typo3 = Request(typo3_probe_url, 'GET')
            try:
                probe_response: Response = await self.crawler.async_send(request_typo3, follow_redirects=False)
            except RequestError:
                self.network_errors += 1
                return False
            if probe_response.status == 200 and "svg" in probe_response.headers.get("content-type", ""):
                return True
        return False


    async def attack(self, request: Request, response: Optional[Response] = None):
        self.finished = True
        request_to_root = Request(request.url)

        if await self.check_typo3(request_to_root.url):
            await self.detect_version(self.PAYLOADS_HASH, request_to_root.url)
            self.versions = sorted(self.versions, key=lambda x: x.split('.')) if self.versions else []

            self.extensions_list = await self.check_typo3_extensions(request_to_root.url)

            if not self.versions:
                list_hashes = []
                list_paths = await fetch_source_files(request.url, self.crawler_configuration)
                for path in list_paths:
                    hash_val, path = await self.get_url_hash(request.url, path)
                    list_hashes.append(hash_val)
                data_hash = self.get_hashes(self.PAYLOADS_HASH)
                self.versions = await get_common_versions(data_hash, list_hashes)

            typo3_detected = {
                "name": "TYPO3",
                "versions": self.versions if self.versions else [],
                "categories": ["CMS TYPO3"],
                "groups": ["Content"]
            }

            log_blue(
                MSG_TECHNO_VERSIONED,
                "TYPO3",
                self.versions
            )

            if self.versions:
                await self.add_info(
                    finding_class=SoftwareVersionDisclosureFinding,
                    request=request_to_root,
                    info=json.dumps(typo3_detected),
                )
            await self.add_info(
                finding_class=SoftwareNameDisclosureFinding,
                request=request_to_root,
                info=json.dumps(typo3_detected),
            )
            if self.extensions_list:
                for ext in self.extensions_list:
                    extension_detected = {
                        "name": ext,
                        "versions": [],
                        "categories": ["TYPO3 extension"],
                        "groups": ['Add-ons']
                    }
                    log_blue(
                        MSG_TECHNO_VERSIONED,
                        ext,
                        []
                    )
                    await self.add_info(
                        finding_class=SoftwareNameDisclosureFinding,
                        request=request,
                        info=json.dumps(extension_detected),
                        response=response
                    )
        else:
            log_blue(MSG_NO_TYPO3)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the Wapiti project (https://wapiti-scanner.github.io)
# Copyright (C) 2024 Cyberwatch
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
import json
from os.path import join as path_join
from typing import Optional
from urllib.parse import urljoin
from datetime import datetime, timezone
from wapitiCore.parsers.html_parser import Html
from httpx import RequestError

from wapitiCore.attack.network_devices.network_device_common import NetworkDeviceCommon, MSG_TECHNO_VERSIONED
from wapitiCore.net import Request
from wapitiCore.net.response import Response
from wapitiCore.definitions.fingerprint import SoftwareNameDisclosureFinding
from wapitiCore.main.log import log_blue, logging

MSG_NO_PALO_ALTO = "No Palo Alto Product Detected"

class ModulePaloAlto(NetworkDeviceCommon):
    """Detect Palo Alto Devices."""
    version = []
    device_name = "Palo Alto GlobalProtect Portal"
    etag_file = "version-table-Etag-palo-alto.txt"


#La détection de version panOS est grandement inspiré de cette application https://github.com/noperator/panos-scanner
    def check_date(self,version_table: dict, dateEtag: datetime.date) -> list:
        versions = [
            version for version, date in version_table.items() if date == dateEtag
        ]
        return versions


    def load_version_table(self) -> dict:
        with open(path_join(self.DATA_DIR, self.etag_file), encoding="utf-8") as versionTableFile:
            entries = [line.strip().split() for line in versionTableFile.readlines()]
        return {
            e[0]: datetime.strptime(" ".join(e[1:]), "%b %d %Y").date()
            for e in entries
        }


    async def check_palo_alto_global_protect_portal(self, url):
        check_list = ['global-protect/login.esp', '']
        for item in check_list:
            full_url = urljoin(url, item)
            request = Request(full_url, 'GET')
            try:
                response: Response = await self.crawler.async_send(request, follow_redirects=False)
            except RequestError:
                self.network_errors += 1
                continue
            page = Html(response.content, full_url)
            div_header = page.soup.find(id="heading")
            if response.is_success:
                if "GlobalProtect Portal" in page.title:
                    return True
                elif div_header and "GlobalProtect Portal" in div_header.text.strip():
                    return True
                elif page.soup.pan_form is not None:
                    return True

        return False

    async def detect_panos_version(self, url):
        check_list = ['global-protect/portal/css/bootstrap.min.css',
            'global-protect/portal/css/latofonts.css',
            'global-protect/portal/js/jquery.min.js',
            'global-protect/portal/css/login.css',
            'global-protect/portal/images/favicon.ico',
            'js/Pan.js',
            'global-protect/portal/images/bg.png'
        ]
        version_table = self.load_version_table()
        for item in check_list:
            full_url = urljoin(url, item)
            request = Request(full_url, 'GET')
            try:
                response: Response = await self.crawler.async_send(request, follow_redirects=False)
            except RequestError:
                self.network_errors += 1
                continue
            fullEtag = response.headers.get("Etag")
            if fullEtag and len(fullEtag) >= 8:
                if '-' in fullEtag:
                    shortenEtag = fullEtag.split('-', 1)[0]
                else:
                    shortenEtag = fullEtag[-8:]
                # https://developer.mozilla.org/fr/docs/Web/HTTP/Reference/Headers/ETag
                if shortenEtag.startswith("W/\""):
                    shortenEtag = shortenEtag[3:]
                try:
                    date = datetime.fromtimestamp(int(shortenEtag, 16), timezone.utc).date()
                except (ValueError, OverflowError, OSError):
                    logging.warning(f"Could not parse ETag value as timestamp: {shortenEtag!r}")
                    continue
                self.version = self.check_date(version_table, date)
                return True
        return False


    async def attack(self, request: Request, response: Optional[Response] = None):
        self.finished = True
        request_to_root = Request(request.url)
        try:
            if await self.check_palo_alto_global_protect_portal(request_to_root.url):
                await self.detect_panos_version(request_to_root.url)

                palo_alto_detected = {
                    "name": self.device_name,
                    "versions": self.version,
                    "categories": ["Network Equipment"],
                    "groups": ["Content"]
                }

                log_blue(
                    MSG_TECHNO_VERSIONED,
                    self.device_name,
                    self.version
                )                

                await self.add_info(
                    finding_class=SoftwareNameDisclosureFinding,
                    request=request_to_root,
                    info=json.dumps(palo_alto_detected),
                )
            else:
                log_blue(MSG_NO_PALO_ALTO)
        except RequestError as req_error:
            self.network_errors += 1
            logging.error(f"Request Error occurred: {req_error}")
